// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:bloc/bloc.dart';
import 'package:equatable/equatable.dart';
import 'package:wyatt_form_bloc/src/enums/enums.dart';
import 'package:wyatt_form_bloc/src/form/form.dart';
import 'package:wyatt_form_bloc/src/validators/form/every_input_validator.dart';

part 'form_data_state.dart';

abstract class FormDataCubit extends Cubit<FormDataState> {
  FormValidator validationStrategy;
  FormData formCopy;

  FormDataCubit({
    required FormData inputs,
    FormValidator validator = const EveryInputValidator(),
  })  : formCopy = inputs.clone(),
        validationStrategy = validator,
        super(FormDataState(data: inputs));

  /// Change value of a field.
  ///
  /// Inputs:
  /// - `field`: The key of the field to change.
  /// - `dirtyValue`: The new value of the field. (Wrapped in a dirty validator)
  void dataChanged<V>(
    String field,
    FormInputValidator<V, ValidationError> dirtyValue,
  ) {
    final form = state.data.clone();

    if (form.contains(field)) {
      form.updateValidator(field, dirtyValue);
    } else {
      throw Exception('Form field $field not found');
    }

    emit(
      state.copyWith(
        data: form,
        status: validationStrategy.validate(form),
      ),
    );
  }

  /// Just validate the form manually. (Useful if you just update the strategy)
  void validate() {
    final form = state.data;
    emit(state.copyWith(status: validationStrategy.validate(form)));
  }

  /// Update entries list.
  ///
  /// Inputs:
  /// - `data`: The new entries list.
  /// - `operation`: The operation to perform on the entries set.
  void updateFormData(
    FormData data, {
    SetOperation operation = SetOperation.replace,
  }) {
    FormData form = data;

    switch (operation) {
      case SetOperation.replace:
        form = data;
        break;
      case SetOperation.difference:
        form = state.data.difference(data);
        break;
      case SetOperation.intersection:
        form = state.data.intersection(data);
        break;
      case SetOperation.union:
        form = state.data.union(data);
        break;
    }

    emit(
      state.copyWith(
        data: form,
        status: validationStrategy.validate(form),
      ),
    );
  }

  /// Reset all form inputs
  void resetForm() {
    emit(FormDataState(data: formCopy));
  }

  /// Submit the form.
  Future<void> submitForm();
}
