// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

part of 'form.dart';

class FormData extends Equatable {
  final List<FormInput> _inputs;

  const FormData(this._inputs);
  const FormData.empty() : this(const <FormInput>[]);

  /// Returns all inputs as a list
  List<FormInput> inputs() => _inputs;

  /// Returns the input for the associated key
  FormInput inputOf(String key) {
    if (contains(key)) {
      return _inputs.firstWhere((input) => input.key == key);
    } else {
      throw Exception('FormInput with key `$key` does not exist in form');
    }
  }

  /// Updates a input (perform a replace at index).
  void updateInput(String key, FormInput input) {
    if (contains(key)) {
      final index = _inputs.indexOf(
        inputOf(key),
      );
      _inputs[index] = input;
    }
  }

  /// Returns all associated validators as a list
  List<FormInputValidator<V, E>> validators<V, E extends ValidationError>() =>
      _inputs
          .map<FormInputValidator<V, E>>(
            (input) => input.validator as FormInputValidator<V, E>,
          )
          .toList();

  /// A [FormInputValidator] represents the value of a single form input field.
  /// It contains information about the [FormInputStatus], value, as well
  /// as validation status.
  T validatorOf<T>(String key) => inputOf(key).validator as T;

  /// Updates validator of a given input. (perform copyWith)
  void updateValidator<V>(
    String key,
    FormInputValidator<V, ValidationError> dirtyValue,
  ) {
    if (contains(key)) {
      final index = _inputs.indexOf(
        inputOf(key),
      );
      _inputs[index] = _inputs[index].copyWith(validator: dirtyValue);
    }
  }

  /// Returns a validation error if the [FormInputValidator] is invalid.
  /// Returns null if the [FormInputValidator] is valid.
  E? errorOf<E extends ValidationError>(String key) =>
      (inputOf(key).validator as FormInputValidator<dynamic, E>).error;

  /// The value of the associated [FormInputValidator]. For example,
  /// if you have a FormInputValidator for FirstName, the value could be 'Joe'.
  V valueOf<V>(String key) =>
      (inputOf(key).validator as FormInputValidator<V, dynamic>).value;

  /// Returns `true` if the [FormInputValidator] is not valid.
  /// Same as `E? errorOf<V, E>(String key) != null`
  bool isNotValid(String key) => !inputOf(key).validator.valid;

  /// Returns all associated metadata as a list
  List<FormInputMetadata<M>> metadata<M>() => _inputs
      .map<FormInputMetadata<M>>(
        (input) => input.metadata as FormInputMetadata<M>,
      )
      .toList();

  /// Returns the metadata associated. With `M` the type of extra data.
  FormInputMetadata<M> metadataOf<M>(String key) =>
      inputOf(key).metadata as FormInputMetadata<M>;

  /// Updates metadata of a given input. (perform copyWith)
  void updateMetadata<M>(String key, FormInputMetadata<M> meta) {
    if (contains(key)) {
      final index = _inputs.indexOf(
        inputOf(key),
      );
      _inputs[index] = _inputs[index].copyWith(metadata: meta);
    }
  }

  /// Check if this contains an input with the given key.
  bool contains(String key) => _inputs.any((input) => input.key == key);

  /// Makes an intersection set operation and returns newly created [FormData]
  FormData intersection(FormData other) {
    final List<FormInput> inputs = <FormInput>[];

    for (final FormInput i in _inputs) {
      if (other.contains(i.key)) {
        inputs.add(i);
      }
    }

    return FormData(inputs);
  }

  /// Makes a difference set operation and returns newly created [FormData]
  FormData difference(FormData other) {
    final List<FormInput> inputs = <FormInput>[];

    for (final FormInput i in other._inputs) {
      if (!contains(i.key)) {
        inputs.add(i);
      }
    }

    for (final FormInput i in _inputs) {
      if (!other.contains(i.key)) {
        inputs.add(i);
      }
    }

    return FormData(inputs);
  }

  /// Makes an union set operation and returns newly created [FormData]
  FormData union(FormData other) {
    final List<FormInput> inputs = <FormInput>[];

    for (final FormInput i in _inputs) {
      inputs.add(i);
    }

    for (final FormInput i in other._inputs) {
      if (!contains(i.key)) {
        inputs.add(i);
      }
    }

    return FormData(inputs);
  }

  /// Deeply copy this.
  FormData clone() => FormData(
        _inputs.map((input) => input.clone()).toList(),
      );

  /// Export this to [Map] format.
  Map<String, dynamic> toMap() {
    final map = <String, dynamic>{};
    for (final input in _inputs) {
      if (input.metadata.export) {
        map[input.name] = input.validator.value;
      }
    }
    return map;
  }

  /// Export this to [String] format.
  String toJson() => jsonEncode(toMap());

  @override
  bool? get stringify => true;

  @override
  List<Object?> get props => _inputs;
}
